import unittest
import numpy as np
import flavio
from flavio.statistics.functions import *

class TestFunctions(unittest.TestCase):
    def test_deltachi2(self):
        self.assertEqual(delta_chi2(3, 1), 9)
        self.assertAlmostEqual(delta_chi2(1, 2), 2.30, delta=0.006)
        self.assertAlmostEqual(delta_chi2(2, 3), 8.03, delta=0.006)
        self.assertAlmostEqual(delta_chi2(3, 2), 11.83, delta=0.006)

    def test_pull(self):
        self.assertEqual(pull(9, 1), 3)
        self.assertAlmostEqual(pull(2.30, 2), 1, delta=0.006)
        self.assertAlmostEqual(pull(8.03, 3), 2, delta=0.006)
        self.assertAlmostEqual(pull(11.83, 2), 3, delta=0.006)
        self.assertAlmostEqual(pull(delta_chi2(4.52, 8), 8), 4.52, places=10)

    def test_cl(self):
        self.assertAlmostEqual(confidence_level(1), 0.6826894921370859, places=10)
        self.assertAlmostEqual(confidence_level(2), 0.9544997361036416, places=10)
        self.assertAlmostEqual(confidence_level(5), 0.9999994266968562, places=10)

    def test_pvalue(self):
        # trivial
        self.assertAlmostEqual(pvalue(1, 1), 1 - 0.6826894921370859, places=10)
        # non-trivial: http://psychclassics.yorku.ca/Fisher/Methods/tabIII.gif
        self.assertAlmostEqual(pvalue(22.362, 13), 0.05, places=6)



import unittest
import numpy as np
from .common import *

wc   = {'C1': 1. + 5.j, 'C2': np.array([1,2,3], dtype=float), 'C3': np.array([1.j]) }
wc_c = {'C1': 1. - 5.j, 'C2': np.array([1,2,3], dtype=float), 'C3': np.array([-1.j]) }
par  = {'a': 3., 'b': 4.+5.j, 'delta':  0.2}
par_c= {'a': 3., 'b': 4.+5.j, 'delta': -0.2}
dict_A   = {'a': 1., 'b': np.array([2,3]), 'c': 2.,}
dict_B   = {'a': 3., 'b': 5.,              'd': 2.,}
dict_C   = {         'b': np.array([8,9]), 'd': 5.,}
dict_sum = {'a': 4., 'b': np.array([15,17]), 'c': 2., 'd': 7.,}

class TestCommon(unittest.TestCase):
    def test_common(self):
        np.testing.assert_equal(conjugate_wc(wc), wc_c)
        np.testing.assert_equal(conjugate_par(par), par_c)
        np.testing.assert_equal(add_dict((dict_A, dict_B, dict_C)), dict_sum)
        np.testing.assert_equal(add_dict([dict_A, dict_B, dict_C]), dict_sum)
        np.testing.assert_equal(add_dict((dict_C, dict_A, dict_B)), dict_sum)



import flavio
from flavio.classes import Observable, Prediction
import numpy as np
from flavio.physics.edms.common import proton_charges

r"""Functions for neutrinoless $\mu - e$ conversion in different target nuclei"""
def CR_mue(wc_obj, par, nucl):
    r"""Conversion rate independent of the target nucleus"""
    mm = par['m_mu']
    scale = flavio.config['renormalization scale']['mudecays']
    #####overlap integrals and other parameters#####
    pc = proton_charges(par, scale)
    GuSp = (pc['gS_u+d'] + pc['gS_u-d']) / 2
    GdSp = (pc['gS_u+d'] - pc['gS_u-d']) / 2
    GsSp = pc['gS_s']
    GuSn = GdSp
    GdSn = GuSp
    GsSn = GsSp
    D   = par['D ' +nucl]*mm**(5/2)
    Sp  = par['Sp '+nucl]*mm**(5/2)
    Vp  = par['Vp '+nucl]*mm**(5/2)
    Sn  = par['Sn '+nucl]*mm**(5/2)
    Vn  = par['Vn '+nucl]*mm**(5/2)
    omega_capt  = par['GammaCapture '+nucl]
    #####Wilson Coefficients######
    #####Conversion Rate obtained from hep-ph/0203110#####
    flavio.citations.register("Kitano:2002mt")
    wc = wc_obj.get_wc('mue', scale, par, nf_out=3)
    prefac = -np.sqrt(2)/par['GF']
    AL = prefac / ( 4 * mm ) * wc['Cgamma_emu']
    AR = prefac / ( 4 * mm ) * wc['Cgamma_mue'].conjugate()
    gRV = {
        'u': prefac * ( wc['CVRR_mueuu'] + wc['CVLR_uumue'] ) / 2,
        'd': prefac * ( wc['CVRR_muedd'] + wc['CVLR_ddmue'] ) / 2,
        's': prefac * ( wc['CVRR_muess'] + wc['CVLR_ssmue'] ) / 2,
    }
    gLV = {
        'u': prefac * ( wc['CVLR_mueuu'] + wc['CVLL_mueuu'] ) / 2,
        'd': prefac * ( wc['CVLR_muedd'] + wc['CVLL_muedd'] ) / 2,
        's': prefac * ( wc['CVLR_muess'] + wc['CVLL_muess'] ) / 2,
    }
    gRS = {
        'u': prefac * ( wc['CSRR_emuuu'] + wc['CSRL_emuuu'] ).conjugate() / 2,
        'd': prefac * ( wc['CSRR_emudd'] + wc['CSRL_emudd'] ).conjugate() / 2,
        's': prefac * ( wc['CSRR_emuss'] + wc['CSRL_emuss'] ).conjugate() / 2,
    }
    gLS = {
        'u': prefac * ( wc['CSRR_mueuu'] + wc['CSRL_mueuu'] ) / 2,
        'd': prefac * ( wc['CSRR_muedd'] + wc['CSRL_muedd'] ) / 2,
        's': prefac * ( wc['CSRR_muess'] + wc['CSRL_muess'] ) / 2,
    }
    lhc = (
        AR.conjugate() * D
        + ( 2 * gLV['u'] + gLV['d'] ) * Vp
        + ( gLV['u'] + 2 * gLV['d'] ) * Vn
        + ( GuSp * gLS['u'] + GdSp * gLS['d'] + GsSp * gLS['s'] ) * Sp
        + ( GuSn * gLS['u'] + GdSn * gLS['d'] + GsSn * gLS['s'] ) * Sn
    )
    rhc = (
        AL.conjugate() * D
        + ( 2 * gRV['u'] + gRV['d'] ) * Vp
        + ( gRV['u'] + 2 * gRV['d'] ) * Vn
        + ( GuSp * gRS['u'] + GdSp * gRS['d'] + GsSp * gRS['s'] ) * Sp
        + ( GuSn * gRS['u'] + GdSn * gRS['d'] + GsSn * gRS['s'] ) * Sn
    )
    omega_conv = 2 * par['GF']**2 * ( abs(lhc)**2 + abs(rhc)**2 )
    return omega_conv / omega_capt


def CR_mueAu(wc_obj, par):
    r"""Conversion rate for $\phantom k^{197}_{79} \mathrm{Au}$"""
    return CR_mue(wc_obj, par, 'Au')
def CR_mueAl(wc_obj, par):
    r"""Conversion rate for $\phantom k^{27}_{13} \mathrm{Al}$"""
    return CR_mue(wc_obj, par, 'Al')
def CR_mueTi(wc_obj, par):
    r"""Conversion rate for $\phantom k^{48}_{22} \mathrm{Ti}$"""
    return CR_mue(wc_obj, par, 'Ti')

CRAu = Observable('CR(mu->e, Au)')
Prediction('CR(mu->e, Au)', CR_mueAu)
CRAu.tex = r"$CR(\mu - e)$ in $\phantom k^{197}_{79} \mathrm{Au}$"
CRAu.description = r"Coherent conversion rate of $\mu^-$ to $e^-$ in $\phantom k^{197}_{79} \mathrm{Au}$"
CRAu.add_taxonomy(r'Process :: muon decays :: LFV decays :: $\mu N \to e N$ :: ' + CRAu.tex)

CRAl = Observable('CR(mu->e, Al)')
Prediction('CR(mu->e, Al)', CR_mueAl)
CRAl.tex = r"$CR(\mu - e)$ in $\phantom k^{27}_{13} \mathrm{Al}$"
CRAl.description = r"Coherent conversion rate of $\mu^-$ to $e^-$ in $\phantom k^{27}_{13} \mathrm{Al}$"
CRAl.add_taxonomy(r'Process :: muon decays :: LFV decays :: $\mu N \to e N$ :: ' + CRAl.tex)

CRTi = Observable('CR(mu->e, Ti)')
Prediction('CR(mu->e, Ti)', CR_mueTi)
CRTi.tex = r"$CR(\mu - e)$ in $\phantom k^{48}_{22} \mathrm{Ti}$"
CRTi.description = r"Coherent conversion rate of $\mu^-$ to $e^-$ in $\phantom k^{48}_{22} \mathrm{Ti}$"
CRTi.add_taxonomy(r'Process :: muon decays :: LFV decays :: $\mu N \to e N$ :: ' + CRTi.tex)



import flavio
from flavio.physics.bdecays.formfactors import hqet
from flavio.physics.bdecays.formfactors import common
from math import sqrt

process_dict = {}
process_dict['B->D'] = {'B': 'B0', 'q': 'b->c', 'P': 'D+'}


def h_to_f(mB, mP, h, q2):
    """Convert HQET form factors to the standard basis.

    See e.g. arXiv:1309.0301, eq. (31)"""
    ff = {}
    r = mP / mB
    ff['f+'] = ((r + 1) * h['h+'] + (r - 1) * h['h-']) / (2 * sqrt(r))
    fminus = ((r - 1) * h['h+'] + (r + 1) * h['h-']) / (2 * sqrt(r))
    ff['f0'] = ff['f+'] + fminus * q2 / (mB**2 - mP**2)
    ff['fT'] = (r + 1) / (2 * sqrt(r)) * h['hT']
    return ff


def ff(process, q2, par, scale, order_z=3, order_z_slp=2, order_z_sslp=1):
    r"""Central value of $B\to P$ form factors in the lattice convention
    CLN parametrization.

    See arXiv:hep-ph/9712417 and arXiv:1703.05330.
    """
    flavio.citations.register("Caprini:1997mu")
    pd = process_dict[process]
    mB = par['m_' + pd['B']]
    mP = par['m_' + pd['P']]
    w = max((mB**2 + mP**2 - q2) / (2 * mB * mP), 1)
    phqet = hqet.get_hqet_parameters(par)
    ash = phqet['ash']
    epsc = phqet['epsc']
    epsb = phqet['epsb']
    zc = phqet['zc']
    CV1 = hqet.CV1(w, zc)
    CV2 = hqet.CV2(w, zc)
    CV3 = hqet.CV3(w, zc)
    CT1 = hqet.CT1(w, zc)
    CT2 = hqet.CT2(w, zc)
    CT3 = hqet.CT3(w, zc)
    z = common.z(mB, mP, q2, t0='tm')
    rho2 = par['CLN rho2_xi']
    c = par['CLN c_xi']
    xi3 = par['CLN xi3']
    xi = hqet.xi(z, rho2, c, xi3, order_z=order_z)
    L = hqet.Lz(par, w, z, order_z=order_z_slp)
    ell = hqet.ell(par, z, order_z=order_z_sslp)
    h = {}
    h['h+'] = xi * (1 + ash * (CV1 + (w + 1) / 2 * (CV2 + CV3))
                    + (epsc + epsb) * L[1]
                    + epsc**2 * ell[1])
    h['h-'] = xi * (ash * (w + 1) / 2 * (CV2 - CV3)
                    + (epsc - epsb) * L[4]
                    + epsc**2 * ell[4])
    h['hT'] = xi * (1 + ash * (CT1 - CT2 + CT3)
                    + (epsc + epsb) * (L[1] - L[4])
                    + epsc**2 * (ell[1] - ell[4]))
    return h_to_f(mB, mP, h, q2)



import unittest
import numpy as np
from .bllgamma import *
from flavio.physics.eft import WilsonCoefficients
from flavio.physics.bdecays.wilsoncoefficients import wctot_dict
from flavio.parameters import default_parameters
import flavio
import copy

c = copy.deepcopy(default_parameters)
c.set_constraint('f_Bs', '0.2303')
c.set_constraint('f_B0', '0.192')
par = c.get_central_all()

wc_obj = WilsonCoefficients()

class TestBllgamma(unittest.TestCase):
    def test_bllgamma(self):
        # numerical comparison to the code used for 1708.02649
        self.assertAlmostEqual(bllg_dbrdq2_int(0.04465, 8.641, wc_obj, par, 'Bs', 'mu')*10**9, 8.4, places=0)
        self.assertAlmostEqual(bllg_dbrdq2_int(15.84, 28.27, wc_obj, par, 'Bs', 'mu')*10**9, 1.7, places=0)
        self.assertAlmostEqual(bllg_dbrdq2_int(0.04465, 8.641, wc_obj, par, 'Bs', 'e')*10**9, 9.0, places=0)
        self.assertAlmostEqual(bllg_dbrdq2_int(15.84, 28.27, wc_obj, par, 'Bs', 'e')*10**9, 1.4, places=0)
        

